#!/usr/bin/env python

# Create IDA IDT files based on an SDK (ARM)

import sys
import os
import string
import types
from romlib import *
import pickle
import commands
import re
from struct import *

def listDirectory(directory, fileExtList):                                        
    "get list of file info objects for files of particular extensions"
    fileList = [os.path.normcase(f)
                for f in os.listdir(directory)]           
    fileList = [f for f in fileList
                if os.path.splitext(f)[1].lower() in fileExtList] 
    return fileList


def getexports(cmd, filename):
    cmd = cmd+" "+filename
    #outtext = commands.getoutput(cmd)
    #print cmd
    ordinal = 0
    name = ""
    section = ""
    exports = {}
    #pipe = os.popen('{ ' + cmd + '; } 2>&1', 'r')
    pipe = os.popen(cmd, 'r')
    line = pipe.readline()
    while line != "":
        #print line
        sectionre = re.search(" of section (\S+)\:",line)
        if (sectionre):
            section = sectionre.groups()[0]
        if (section == ".text"):
            namere = re.search("00000000 \<(.+)\>\:",line)
            if (namere):
                name = namere.groups()[0]
                #print "%d: %s"%(ordinal, name)
                exports[ordinal] = name
        if (section == ".idata$5"):
            ordre = re.search("^\s+0000\s+(\S+)",line)
            if (ordre):
                ordinal, = unpack(">L",pack("<L",int(ordre.groups()[0],16)))
                ordinal &= 0x7fffffff
        line = pipe.readline()
    return exports

def sortByNumber(a, b):
    return (a > b)

# Main
if __name__ == '__main__':
    if len(sys.argv) < 2:
        print
        print 'usage: %s sdk-dir <objdump-prefix>' % sys.argv[0]
        print
        print 'sdk-dir is the directory where the *.lib files can be found.'
        print 'resulting *.idt files will be stored in a subdir called idt'
        print 'in the current directory. If it exists it will be deleted.'
        print
        print '<objdump-prefix> is an optional prefix for the ARM objdump'
        print 'For example arm-elf in the case of arm-elf-objdump'
        sys.exit()
    else:
        sdkdir = sys.argv[1]

    if len(sys.argv) == 3:
        prefix = sys.argv[2] + "-"
    else:
        prefix = ""
        
    try:
        os.mkdir('idt')
    except OSError, inst:
        if inst.errno == 17:
            for root, dirs, files in os.walk('idt', topdown=False):
                for name in files:
                    os.remove(os.path.join(root,name))
                for name in dirs:
                    os.rmdir(os.path.join(root, name))

    for i in listDirectory(sdkdir, [".lib"]):
        print i
        cmd = prefix+"objdump -s --disassemble-all "
        exportfuncs = getexports(cmd, sdkdir + os.sep + i)
        exportfuncs.keys().sort(sortByNumber)
        filename = os.path.splitext(i.lower())[0]+".idt"
        idtfile = open('idt/'+filename,'w')
        print >> idtfile, "0\tName=%s"%(os.path.splitext(i.lower())[0]+".dll")
        for ord in exportfuncs.keys():
            funcname = os.path.splitext(i.lower())[0]+"_"+str(ord)
            print >> idtfile, "%d\tName=%s"%(ord,exportfuncs[ord])
        idtfile.close

