#!/usr/bin/env python

# Build data structures representing all information about files in the ROM

import sys
import os
from mmap import *
import struct, sstruct
import string
import types
import codecs
from romlib import *
import pickle

exports = ExportBase()

importedbyaddr = {}
importedbyname = {}
entrypoints = {}
filebyuid = {}

def myuniq(alist):
    set = {}
    map(set.__setitem__, alist, [])
    return set.keys()

def processdir(romdir):
    for file in romdir.dir:
        if file.iAtt & 0x10:
            processdir(file.romdir)
        else:
            if ( file.fileheader.iUid1 == 0x10000079) | ( file.fileheader.iUid1 == 0x1000007A):
                filebyuid[file.fileheader.iUid3] = file.fullname()
                entrypoints[file.fileheader.iEntryPoint] = file.name
                # an exception for these files as they result in conflicts with elocl.dll
                if ((file.name.lower() == "elocl.loc") | (file.name.lower() == "elocl.01")):
                    importedbyname[file.name.lower()]=[]
                    continue
                if (file.fileheader.iExportDirCount != 0):
                    for export in range(file.fileheader.iExportDirCount):
                        # create an export entry indexed by the export table address
                        exports.byaddr[file.fileheader.iExportDir+export*4] = ExportEntry(file.fileheader.iExportDir+export*4,\
                                                                                file.fileheader.iExports[export],\
                                                                                export+1,\
                                                                                file.name,\
                                                                                file.fullname(),\
                                                                                os.path.splitext(file.name.lower())[0]+'_'+str(export+1))
                        # create an export entry indexed by the symbolic exportname
                        exports.byfunc[exports.byaddr[file.fileheader.iExportDir+export*4].funcname] =  exports.byaddr[file.fileheader.iExportDir+export*4];
                        # initialize an empty list for enumerating all files using this export
                        importedbyaddr[file.fileheader.iExportDir+export*4]=[]
                    # initialize an empty list for this to add the files that import this file later
                    importedbyname[file.name.lower()]=[]
            else:
                filebyuid[file.fileheader.iUid1] = file.fullname()

def resolvimports(romdir):
    for file in romdir.dir:
        if file.iAtt & 0x10:
            resolvimports(file.romdir)
        else:
            if ( file.fileheader.iUid1 == 0x10000079) | ( file.fileheader.iUid1 == 0x1000007A):
                if (file.fileheader.iDllRefTable != 0):
                    for i in range(file.dll.iNumberOfEntries):
                        # add this file to the 'imported by' list for the referenced dll
                        if entrypoints.has_key(file.dll.DllEntries[i].iEntryPoint):
                            importedbyname[entrypoints[file.dll.DllEntries[i].iEntryPoint].lower()].append(file.fullname())

                    # search for THUMB stub
                    sub = struct.pack("<HHHH", 0x4b01, 0x681b, 0x4718, 0x46c0);
                    data = file.file[file.fileheader.address:file.fileheader.end+file.iSize]
                    n = data.find(sub)
                    while n>0:
                        addr, = struct.unpack("<L", data[n+8:n+12]);
                        if (importedbyaddr.has_key(addr)):
                            importedbyaddr[addr].append(file.fullname());
                        else:
                            print "cannot find %08x for %s"%(addr,file.name)
                        n = data.find(sub,n+12)

                    # search for ARM stub
                    sub = struct.pack("<LL", 0xE59FC000, 0xE59Cf000);
                    n = data.find(sub)
                    while n>0:
                        addr, = struct.unpack("<L", data[n+8:n+12]);
                        if (importedbyaddr.has_key(addr)):
                            importedbyaddr[addr].append(file.fullname());
                        else:
                            print "cannot find %08x for %s"%(addr,file.name)
                        n = data.find(sub,n+12)


def buildrominfo(romdir):
    processdir(romdir)
    resolvimports(romdir)

def displaydir(romdir):
    for file in romdir.dir:
        if file.iAtt & 0x10:
            print 'dir: %10d 0x%0.8x 0x%0.2x %s %s'%(file.iSize,file.iAddressLin,file.iAtt,file.path(),file.name)
            displaydir(file.romdir)
        else:
            if ( file.fileheader.iUid1 == 0x10000079) | ( file.fileheader.iUid1 == 0x1000007A):
#                print '\tiUids: 0x%0x, 0x%0x, 0x%0x'%(file.fileheader.iUid1,file.fileheader.iUid2,file.fileheader.iUid3)
                print '\t%10d 0x%0.8x 0x%0.2x %s'%(file.iSize,file.iAddressLin,file.iAtt,file.fullname())
#                if (getattr(file.filheaderm, "iDllRefTable", None) != None & file.fileheader.iDllRefTable != 0):
                if (file.fileheader.iDllRefTable != 0):
#                    print 'DllRefTable 0x%08x'%file.fileheader.iDllRefTable
                    print '\t\tiNumberOfEntries: = %d'%file.dll.iNumberOfEntries
                    for i in range(file.dll.iNumberOfEntries):
                        if entrypoints.has_key(file.dll.DllEntries[i].iEntryPoint):
                            print '\t\tdepends on: %s (0x%08x)'% (entrypoints[file.dll.DllEntries[i].iEntryPoint],file.dll.DllEntries[i].iEntryPoint)
                        else:
                            print '\t\tdepends on: <UNKNOWN> (0x%08x)'%file.dll.DllEntries[i].iEntryPoint


# Main
if __name__ == '__main__':
    if len(sys.argv) < 2:
        print 'usage: %s filename' % sys.argv[0]
        sys.exit()
    else:
        rom= Rom(sys.argv[1])

    print 'Building a ROM information db based on %s:'%sys.argv[1]
    print 'iRomBase:\t\t0x%0.8x' % rom.iRomBase
    print 'iRomSize:\t\t0x%0.8x' % rom.iRomSize
    print 'iRomRootDirectoryList:\t0x%0.8x' % rom.iRomRootDirectoryList
    print 'iPrimaryFile:\t\t0x%0.8x' % rom.iPrimaryFile
    print 'iSecondaryFile:\t\t0x%0.8x' % rom.iSecondaryFile
    print 'iKernDataAddress:\t0x%0.8x' % rom.iKernDataAddress
    print 'iKernStackAddress:\t0x%0.8x' % rom.iKernStackAddress
    print 'iRomSectionHeader:\t0x%0.8x' % rom.iRomSectionHeader
    print 'iVariantFile:\t\t0x%0.8x' % rom.iVariantFile
    print 'iExtensionFile:\t\t0x%0.8x' % rom.iExtensionFile
    print 'iRelocInfo:\t\t0x%0.8x' % rom.iRelocInfo
    print 'iUserDataAddress:\t0x%0.8x' % rom.iUserDataAddress
    print 'iTotalUserDataSize:\t0x%0.8x' % rom.iTotalUserDataSize

    rootdirlist = RootDirList(rom, rom.iRomRootDirectoryList);
    print 'iNumRootDirs: 0x%0.8x' % rootdirlist.iNumRootDirs

    buildrominfo(rootdirlist.rootdirinfo.rootdir)
    #displaydir(rootdirlist.rootdirinfo.rootdir)
    #for name in importedbyname['ezlib.dll']:
    #    print 'name: %s'%name
    #print 'uid: 0x%08x, file: %s'%(0x101FD5D8, filebyuid[0x101FD5D8])

    #keys = importedbyaddr.keys();
    #for i in keys:
    #    print "0x%08x"%(i)
    #    print len(importedbyaddr[i])
    #    print importedbyaddr[i]

    pickle.dump(filebyuid,open('filebyuid.db','w'))
    pickle.dump(importedbyaddr,open('importedbyaddr.db','w'))
    pickle.dump(importedbyname,open('importedbyname.db','w'))
    pickle.dump(exports,open('exports.db','w'))

    rom.file.close()
